@@ -1,4 +1,4 @@
-# Copyright (c) 2015-2022 Patricio Cubillos and contributors.
+# Copyright (c) 2015-2023 Patricio Cubillos and contributors.
 # mc3 is open-source software under the MIT license (see LICENSE).
 
 __all__ = [
@@ -20,22 +20,24 @@ import matplotlib.pyplot as plt
 from .fit_driver import fit
 from .mcmc_driver import mcmc
 from .ns_driver import nested_sampling
-from . import utils   as mu
-from . import stats   as ms
-from . import plots   as mp
-from .VERSION import __version__
+from . import utils as mu
+from . import stats as ms
+from . import plots as mp
+from .version import __version__
 
 
-@mu.ignore_system_exit
 def sample(
-    data=None, uncert=None, func=None, params=None, indparams=[],
+    data=None, uncert=None, func=None, params=None,
+    indparams=[], indparams_dict={},
     pmin=None, pmax=None, pstep=None,
     prior=None, priorlow=None, priorup=None,
     sampler=None, ncpu=None, leastsq=None, chisqscale=False,
     nchains=7, nsamples=None, burnin=0, thinning=1,
     grtest=True, grbreak=0.0, grnmin=0.5, wlike=False,
     fgamma=1.0, fepsilon=0.0, hsize=10, kickoff='normal',
-    plots=False, ioff=False, showbp=True, savefile=None, resume=False,
+    plots=False, theme='blue', statistics='med_central',
+    ioff=False, showbp=True,
+    savefile=None, resume=False,
     rms=False, log=None, pnames=None, texnames=None,
     **kwargs):
     """
@@ -49,7 +51,7 @@ def sample(
         Uncertainties of data.
     func: Callable or string-iterable
         The callable function that models data as:
-            model = func(params, *indparams)
+            model = func(params, *indparams, **indparams_dict)
         Or an iterable of 3 strings (funcname, modulename, path)
         that specifies the function name, function module, and module path.
         If the module is already in the python-path scope, path can be omitted.
@@ -59,6 +61,8 @@ def sample(
     indparams: tuple or string
         Additional arguments required by func.  If string, path to file
         containing indparams.
+    indparams_dict: dict
+        Additional keyword arguments required by func (if needed).
     pmin: 1D ndarray
         Lower boundaries for the posterior exploration.
     pmax: 1D ndarray
@@ -115,7 +119,7 @@ def sample(
         If True, run Gelman & Rubin test.
     grbreak: Float
         Gelman-Rubin convergence threshold to stop the MCMC (I'd suggest
-        grbreak ~ 1.001--1.005).  Do not break if grbreak=0.0 (default).
+        grbreak ~ 1.01).  Do not break if grbreak=0.0 (default).
     grnmin: Integer or float
         Minimum number of samples required for grbreak to stop the MCMC.
         If grnmin > 1: grnmin sets the minimum required number of samples.
@@ -135,6 +139,14 @@ def sample(
     plots: Bool
         If True plot parameter traces, pairwise-posteriors, and posterior
         histograms.
+    theme:
+        The color theme for plots. Can have any format recognized as a
+        matplotlib color.
+    statistics: String
+        Statistics to adopt for the plots. Select from:
+        - 'med_central' Median and central quantile
+        - 'max_like' Max marginal likelihood (mode) and HPD
+        - 'global_max_like' Max a posteriori (best-fit) and HPD
     ioff: Bool
         If True, set plt.ioff(), i.e., do not display figures on screen.
     showbp: Bool
@@ -271,13 +283,13 @@ def sample(
        f"{log.sep}\n\n")
 
     if sampler is None:
-        log.error("'sampler' is a required argument.")
+        log.error("'sampler' is a required argument")
     if nsamples is None and sampler in ['MRW', 'DEMC', 'snooker']:
-        log.error("'nsamples' is a required argument for MCMC runs.")
+        log.error("'nsamples' is a required argument for MCMC runs")
     if leastsq not in [None, 'lm', 'trf']:
         log.error(
             f"Invalid 'leastsq' input ({leastsq}). Must select from "
-             "['lm', 'trf'].")
+             "['lm', 'trf']")
 
     # Read the model parameters:
     params = mu.isfile(params, 'params', log, 'ascii', False, not_none=True)
@@ -294,7 +306,7 @@ def sample(
             pmin = params[1]
             pmax = params[2]
         else:
-            log.error('Invalid format/shape for params input file.')
+            log.error('Invalid format/shape for params input file')
         params = params[0]     # The initial guess
     params = np.array(params)
 
@@ -304,7 +316,7 @@ def sample(
         data, uncert = data
     # Make local 'uncert' a copy, to avoid overwriting:
     if uncert is None:
-        log.error("'uncert' is a required argument.")
+        log.error("'uncert' is a required argument")
     uncert = np.copy(uncert)
 
     # Process the independent parameters:
@@ -329,7 +341,7 @@ def sample(
     elif not callable(func):
         log.error(
             "'func' must be either a callable or an iterable of strings "
-            "with the model function, file, and path names.")
+            "with the model function, file, and path names")
 
     if ncpu is None and sampler in ['snooker', 'demc', 'mrw']:
         ncpu = nchains
@@ -364,7 +376,7 @@ def sample(
     pmax = np.asarray(pmax)
     if (np.any(np.isinf(pmin)) or np.any(np.isinf(pmax))) \
             and sampler=='dynesty':
-        log.error('Parameter space must be constrained by pmin and pmax.')
+        log.error('Parameter space must be constrained by pmin and pmax')
 
     if pstep is None:
         pstep = 0.1 * np.abs(params)
@@ -391,18 +403,20 @@ def sample(
             "Some initial-guess values are out of bounds:\n"
             "Param name           pmin          value           pmax\n"
             "-----------  ------------   ------------   ------------"
-           f"{pout}")
+            f"{pout}"
+        )
 
     nfree = int(np.sum(pstep > 0))
     ifree = np.where(pstep > 0)[0]  # Free parameter indices
     ishare = np.where(pstep < 0)[0]  # Shared parameter indices
 
     # Check output dimension:
-    model0 = func(params, *indparams)
+    model0 = func(params, *indparams, **indparams_dict)
     if np.shape(model0) != np.shape(data):
         log.error(
             f"The size of the data array ({np.size(data)}) does not "
-            f"match the size of the func() output ({np.size(model0)}).")
+            f"match the size of the func() output ({np.size(model0)})"
+        )
 
     # Check that output path exists:
     if savefile is not None:
@@ -410,7 +424,8 @@ def sample(
         if not os.path.exists(fpath):
             log.warning(
                 f"Output folder path: '{fpath}' does not exist. "
-                "Creating new folder.")
+                "Creating new folder."
+            )
             os.makedirs(fpath)
 
     # At the moment, skip optimization when these dynesty inputs exist:
@@ -422,7 +437,8 @@ def sample(
     chisq_factor = 1.0
     if leastsq is not None:
         fit_output = fit(
-            data, uncert, func, np.copy(params), indparams,
+            data, uncert, func, np.copy(params),
+            indparams, indparams_dict,
             pstep, pmin, pmax, prior, priorlow, priorup, leastsq)
         fit_bestp = fit_output['bestp']
         log.msg(
@@ -436,7 +452,8 @@ def sample(
 
             # Re-calculate best-fitting parameters with new uncertainties:
             fit_output = fit(
-                data, uncert, func, np.copy(params), indparams,
+                data, uncert, func, np.copy(params),
+                indparams, indparams_dict,
                 pstep, pmin, pmax, prior, priorlow, priorup, leastsq)
             log.msg(
                 "Least-squares best-fitting parameters (rescaled chisq):"
@@ -453,18 +470,29 @@ def sample(
 
     # Here's where the magic happens:
     if sampler in ['mrw', 'demc', 'snooker']:
-        output = mcmc(data, uncert, func, params, indparams, pmin, pmax, pstep,
+        output = mcmc(
+            data, uncert, func,
+            params, indparams, indparams_dict,
+            pmin, pmax, pstep,
             prior, priorlow, priorup, nchains, ncpu, nsamples, sampler,
             wlike, fit_output, grtest, grbreak, grnmin, burnin, thinning,
-            fgamma, fepsilon, hsize, kickoff, savefile, resume, log)
+            fgamma, fepsilon, hsize, kickoff, savefile, resume, log,
+            pnames, texnames,
+        )
     elif sampler == 'dynesty':
-        output = nested_sampling(data, uncert, func, params, indparams,
+        output = nested_sampling(
+            data, uncert, func, params, indparams, indparams_dict,
             pmin, pmax, pstep, prior, priorlow, priorup, ncpu,
-            thinning, resume, log, **kwargs)
+            thinning, resume, log, **kwargs,
+        )
+
+    # Get some stats:
+    output['chisq_factor'] = chisq_factor
 
     if leastsq is not None:
-        if (output['best_log_post'] - fit_output['best_log_post'] > 5.0e-8
-            and np.any((output['bestp'] - fit_output['bestp'])!=0.0)):
+        delta_log_post = output['best_log_post'] - fit_output['best_log_post']
+        delta_pars = output['bestp'] - fit_output['bestp']
+        if delta_log_post > 5.0e-8 and np.any(delta_pars != 0.0):
             log.warning(
                 "MCMC found a better fit than the minimizer:\n"
                 "MCMC best-fitting parameters:        (chisq={:.8g})\n{}\n"
@@ -472,101 +500,96 @@ def sample(
                 format(
                     -2*output['best_log_post'], output['bestp'],
                     -2*fit_output['best_log_post'], fit_output['bestp']))
-        else:
-            output['best_log_post'] = fit_output['best_log_post']
-            output['best_chisq'] = fit_output['best_chisq']
-            output['best_model'] = fit_output['best_model']
-            output['bestp'] = fit_output['bestp']
 
-    # And remove burn-in samples:
+    # Stats without burn-in samples:
     posterior, zchain, zmask = mu.burn(
         Z=output['posterior'], zchain=output['zchain'], burnin=output['burnin'])
 
-    # Get some stats:
-    output['chisq_factor'] = chisq_factor
-    output['BIC'] = output['best_chisq'] + nfree*np.log(ndata)
-    if ndata > nfree:
-        output['red_chisq'] = output['best_chisq']/(ndata-nfree)
-    else:
-        output['red_chisq'] = np.nan
-    output['stddev_residuals'] = np.std(output['best_model']-data)
-
-    # Compute the credible region for each parameter:
     bestp = output['bestp']
-    CRlo = np.zeros(nparams)
-    CRhi = np.zeros(nparams)
-    pdf  = []
-    xpdf = []
-    for post, idx in zip(posterior.T, ifree):
-        PDF, Xpdf, HPDmin = ms.cred_region(post)
-        pdf.append(PDF)
-        xpdf.append(Xpdf)
-        CRlo[idx] = np.amin(Xpdf[PDF>HPDmin])
-        CRhi[idx] = np.amax(Xpdf[PDF>HPDmin])
-    # CR relative to the best-fitting value:
-    CRlo[ifree] -= bestp[ifree]
-    CRhi[ifree] -= bestp[ifree]
+    post = mp.Posterior(
+        posterior, pnames=texnames[ifree], theme=theme,
+        bestp=bestp[ifree], statistics=statistics,
+    )
+    print(post.statistics)
+    # Let Posterior to turn the theme into a Theme() object:
+    theme = post.theme
 
-    # Get the mean and standard deviation from the posterior:
-    meanp = np.zeros(nparams, np.double) # Parameters mean
-    stdp  = np.zeros(nparams, np.double) # Parameter standard deviation
-    meanp[ifree] = np.mean(posterior, axis=0)
-    stdp [ifree] = np.std(posterior,  axis=0)
-    for s in ishare:
-        bestp[s] = bestp[-int(pstep[s])-1]
-        meanp[s] = meanp[-int(pstep[s])-1]
-        stdp [s] = stdp [-int(pstep[s])-1]
-        CRlo [s] = CRlo [-int(pstep[s])-1]
-        CRhi [s] = CRhi [-int(pstep[s])-1]
-
-    output['CRlo'] = CRlo
-    output['CRhi'] = CRhi
-    output['stdp'] = stdp
-    output['meanp'] = meanp
-    output['ifree'] = ifree
-    output['pnames'] = pnames
-    output['texnames'] = texnames
+    # Parameter statistics:
+    sample_stats = ms.calc_sample_statistics(
+        post.posterior, bestp, pstep, calc_hpd=True,
+    )
+    median = output['medianp'] = sample_stats[0]
+    mean = output['meanp'] = sample_stats[1]
+    stdp = output['stdp'] = sample_stats[2]
+    med_low_bounds = output['median_low_bounds'] = sample_stats[3]
+    med_high_bounds = output['median_high_bounds'] = sample_stats[4]
+    mode = output['mode'] = sample_stats[5]
+    hpd_low_bounds = output['hpd_low_bounds'] = sample_stats[6]
+    hpd_high_bounds = output['hpd_high_bounds'] = sample_stats[7]
+    # Legacy (this will be deprecated at some point)
+    output['CRlo'] = hpd_low_bounds - bestp
+    output['CRhi'] = hpd_high_bounds - bestp
+    output['CRlo'][pstep==0] = output['CRhi'][pstep==0] = 0.0
 
     log.msg(
-        "\nParam name     Best fit   Lo HPD CR   Hi HPD CR        Mean    Std dev       S/N"
-        "\n----------- ----------------------------------- ---------------------- ---------",
+        "\nParameter name     best fit   median      1sigma_low   1sigma_hi        S/N"
+        "\n--------------- -----------  -----------------------------------  ---------",
         width=80)
     for i in range(nparams):
-        snr = f"{np.abs(bestp[i])/stdp[i]:.1f}"
-        mean = f"{meanp[i]: 11.4e}"
-        lo = f"{CRlo[i]: 11.4e}"
-        hi = f"{CRhi[i]: 11.4e}"
-        if i in ifree:  # Free-fitting value
-            pass
-        elif i in ishare: # Shared value
-            snr = f"[share{-int(pstep[i]):02d}]"
-        else:             # Fixed value
+        pname = f'{pnames[i][0:15]:<15}'
+        lo = med_low_bounds[i] - median[i]
+        hi = med_high_bounds[i] - median[i]
+        if i in ifree:
+            snr = f"{np.abs(bestp[i])/stdp[i]:.1f}"
+        elif i in ishare:
+            idx = -int(pstep[i])
+            snr = f"[share{idx:02d}]"
+        else:
             snr = "[fixed]"
-            mean = f"{bestp[i]: 11.4e}"
+            lo = hi = 0.0
         log.msg(
-            f"{pnames[i][0:11]:<11s} {bestp[i]:11.4e} {lo:>11s} {hi:>11s} "
-            f"{mean:>11s} {stdp[i]:10.4e} {snr:>9s}",
-            width=160)
+            f"{pname} {bestp[i]:11.4e}  {median[i]:11.4e} "
+            f"{lo:11.4e} {hi:11.4e}  {snr:>9s}",
+            width=160,
+        )
 
-    fmt = len(f"{output['BIC']:.4f}")  # Length of string formatting
-    log.msg(" ")
-    if chisqscale:
-        log.msg(
-            "sqrt(reduced chi-squared) factor:   {:{}.4f}".
-            format(output['chisq_factor'], fmt), indent=2)
-    log.msg("Best-parameter's chi-squared:       {:{}.4f}".
-            format(output['best_chisq'], fmt), indent=2)
-    log.msg("Best-parameter's -2*log(posterior): {:{}.4f}".
-            format(-2*output['best_log_post'], fmt), indent=2)
-    log.msg("Bayesian Information Criterion:     {:{}.4f}".
-            format(output['BIC'], fmt), indent=2)
-    log.msg("Reduced chi-squared:                {:{}.4f}".
-            format(output['red_chisq'], fmt), indent=2)
-    log.msg("Standard deviation of residuals:  {:.6g}\n".
-            format(output['stddev_residuals']), indent=2)
+    # Fit statistics:
+    best_chisq = output['best_chisq']
+    log_post = -2.0*output['best_log_post']
+    bic = output['BIC']
+    red_chisq = output['red_chisq']
+    std_dev = output['stddev_residuals']
 
-    if savefile is not None or plots or closelog:
-        log.msg("\nOutput sampler files:")
+    chisqscale_txt = f"sqrt(reduced chi-squared) factor: {chisq_factor:.4f}\n"
+    if not chisqscale:
+        chisqscale_txt = ''
+
+    fmt = len(f"{bic:.4f}")  # Length of string formatting
+    log.msg(
+        f"\n{chisqscale_txt}"
+        f"Best-parameter's chi-squared:       {best_chisq:{fmt}.4f}\n"
+        f"Best-parameter's -2*log(posterior): {log_post:{fmt}.4f}\n"
+        f"Bayesian Information Criterion:     {bic:{fmt}.4f}\n"
+        f"Reduced chi-squared:                {red_chisq:{fmt}.4f}\n"
+        f"Standard deviation of residuals:  {std_dev:.6g}\n",
+        indent=2,
+    )
+
+    # Extract filename from savefile or default one:
+    if savefile is not None:
+        savefile_root = os.path.splitext(savefile)[0]
+    else:
+        savefile_root = 'mc3'
+
+    stats_file = f'{savefile_root}_statistics.txt'
+    ms.summary_stats(post, output, filename=stats_file)
+    log.msg(
+        '\nFor a detailed summary with all parameter posterior statistics '
+        f'see {stats_file}',
+    )
+
+    log.msg("\nOutput sampler files:")
+    log.msg(stats_file, indent=2)
 
     # Save results (pop unpickables before saving, then put back):
     if savefile is not None:
@@ -575,54 +598,40 @@ def sample(
         tmp_outputs = {key: output.pop(key) for key in unpickables}
         np.savez(savefile, **output)
         output.update(tmp_outputs)
-        log.msg(f"'{savefile}'", indent=2)
+        log.msg(savefile, indent=2)
 
     if plots:
-        # Extract filename from savefile or default to sampler:
-        fname = sampler if savefile is None else os.path.splitext(savefile)[0]
-        # Include bestp in posterior plots:
-        best_freepars = output['bestp'][ifree] if showbp else None
-
         # Trace plot:
+        savefile = f'{savefile_root}_trace.png'
         mp.trace(
             output['posterior'], zchain=output['zchain'],
             burnin=output['burnin'], pnames=texnames[ifree],
-            savefile=fname+"_trace.png")
-        log.msg(f"'{fname}_trace.png'", indent=2)
+            savefile=savefile, color=theme.color,
+        )
+        log.msg(savefile, indent=2)
         # Pairwise posteriors:
-        mp.pairwise(
-            posterior, pnames=texnames[ifree], bestp=best_freepars,
-            savefile=fname+"_pairwise.png")
-        log.msg(f"'{fname}_pairwise.png'", indent=2)
+        savefile = f'{savefile_root}_pairwise_posterior.png'
+        post.plot(savefile=savefile, show_estimates=showbp)
+        log.msg(savefile, indent=2)
         # Histograms:
-        mp.histogram(
-            posterior, pnames=texnames[ifree], bestp=best_freepars,
-            savefile=fname+"_posterior.png",
-            quantile=0.683, pdf=pdf, xpdf=xpdf)
-        log.msg(f"'{fname}_posterior.png'", indent=2)
+        savefile = f'{savefile_root}_marginal_posterior.png'
+        post.plot_histogram(savefile=savefile, show_estimates=showbp)
+        log.msg(savefile, indent=2)
         # RMS vs bin size:
         if rms:
-            RMS, RMSlo, RMShi, stderr, bs = ms.time_avg(
-                output['best_model']-data)
+            savefile = f'{savefile_root}_RMS.png'
+            residuals = output['best_model'] - data
+            data_rms, rms_lo, rms_hi, stderr, binsize = ms.time_avg(residuals)
             mp.rms(
-                bs, RMS, stderr, RMSlo, RMShi, binstep=len(bs)//500+1,
-                savefile=fname+"_RMS.png")
-            log.msg(f"'{fname}_RMS.png'", indent=2)
-        # Guessing that indparams[0] is the X array for data as in y=y(x):
-        if (indparams != []
-                and isinstance(indparams[0], (list, tuple, np.ndarray))
-                and np.size(indparams[0]) == ndata):
-            try:
-                mp.modelfit(
-                    data, uncert, indparams[0], output['best_model'],
-                    savefile=fname+"_model.png")
-                log.msg(f"'{fname}_model.png'", indent=2)
-            except:
-                pass
+                binsize, data_rms, stderr, rms_lo, rms_hi,
+                binstep=len(binsize)//500+1,
+                savefile=savefile,
+            )
+            log.msg(savefile, indent=2)
 
     # Close the log file if necessary:
     if closelog:
-        log.msg(f"'{log.logname}'", indent=2)
+        log.msg(log.logname, indent=2)
         log.close()
 
     return output
